{
    "name": "HTML Formatter",
    "type": ["formatter"],
    "requires_setup": False,
    "dependencies": {"python": ["hash_enricher", "loguru", "jinja2"], "bin": [""]},
    "configs": {
        "detect_thumbnails": {
            "default": True,
            "help": "if true will group by thumbnails generated by thumbnail enricher by id 'thumbnail_00'",
            "type": "bool",
        },
    },
    "description": """ """,
}
